package org.processmining.plugins.log.ui.logdialog;

import java.awt.Color;
import java.awt.Dimension;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.ScrollPaneConstants;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;

import com.fluxicon.slickerbox.ui.SlickerScrollBarUI;

public class LogAttributesUI extends JPanel implements TreeSelectionListener {

	private static final long serialVersionUID = 1L;

	private static final String TREE_ROOT_NODE = "Log";
	private static final String TREE_NODE_ATTRIBUTES = "Attributes";
	private static final String TREE_NODE_CLASSIFIERS = "Classifiers";
	private static final String TREE_NODE_GLOBAL_EVENT_ATTRIBUTES = "Global Event Attributes";
	private static final String TREE_NODE_GLOBAL_TRACE_ATTRIBUTES = "Global Trace Attributes";
	private static final String TREE_NODE_EXTENSIONS = "Extensions";

	private XLog log;
	private JTextArea valueArea;

	public LogAttributesUI(XLog log) {
		this.log = log;

		setOpaque(false);
		setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		setLayout(new BoxLayout(this, BoxLayout.X_AXIS));

		// Create the root node of the tree.
		DefaultMutableTreeNode root = new DefaultMutableTreeNode(TREE_ROOT_NODE);

		// Create metadata tree.
		JTree mdtree = new JTree(root);
		mdtree.setBackground(LogPreviewUI.COLOR_ENCLOSURE_BG);
		mdtree.setForeground(Color.WHITE);
		DefaultTreeCellRenderer renderer = (DefaultTreeCellRenderer) mdtree
				.getCellRenderer();
		renderer.setBackground(LogPreviewUI.COLOR_ENCLOSURE_BG);
		renderer
				.setBackgroundNonSelectionColor(LogPreviewUI.COLOR_ENCLOSURE_BG);
		renderer
				.setBackgroundSelectionColor(LogPreviewUI.COLOR_LIST_SELECTION_BG);
		renderer.setForeground(Color.WHITE);
		renderer.setTextNonSelectionColor(Color.WHITE);
		renderer.setTextSelectionColor(LogPreviewUI.COLOR_LIST_SELECTION_FG);

		// Add selection listener so we can react by showing additional
		// information.
		mdtree.addTreeSelectionListener(this);

		// Create scroll pane for the metadata tree.
		JScrollPane mdtreeScrollPane = new JScrollPane(mdtree);
		mdtreeScrollPane.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		mdtreeScrollPane.setOpaque(false);
		mdtreeScrollPane
				.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
		mdtreeScrollPane
				.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		JScrollBar vBar = mdtreeScrollPane.getVerticalScrollBar();
		vBar.setUI(new SlickerScrollBarUI(vBar,
				LogPreviewUI.COLOR_ENCLOSURE_BG, new Color(160, 160, 160),
				LogPreviewUI.COLOR_NON_FOCUS, 4, 12));
		JScrollBar hBar = mdtreeScrollPane.getHorizontalScrollBar();
		hBar.setUI(new SlickerScrollBarUI(hBar,
				LogPreviewUI.COLOR_ENCLOSURE_BG, new Color(160, 160, 160),
				LogPreviewUI.COLOR_NON_FOCUS, 4, 12));

		// Adding all nodes to the tree.

		/*
		 * In the current implementation, the TreeNode elements get a String as
		 * user data. This is normally not the best choice (limited), but, since
		 * keys in XES attributes are also strings and they are the identifying
		 * element for attributes in XES, I do not expect too much trouble.
		 */

		// Build up list of extensions.
		DefaultMutableTreeNode extensions = new DefaultMutableTreeNode(new TreeEntry(TREE_NODE_EXTENSIONS));
		for (XExtension extension : log.getExtensions()) {
			extensions.add(new DefaultMutableTreeNode(new TreeEntry(extension.getName(), extension.getPrefix() + " (" + extension.getName() + ")")));
		}
		root.add(extensions);

		// Build up list of Global Trace Attributes
		DefaultMutableTreeNode traceAttributes = new DefaultMutableTreeNode(new TreeEntry(TREE_NODE_GLOBAL_TRACE_ATTRIBUTES));
		for (XAttribute attribute : log.getGlobalTraceAttributes()) {
			traceAttributes.add(new DefaultMutableTreeNode(new TreeEntry(attribute.getKey())));
		}
		root.add(traceAttributes);

		// Build up list of Global Event Attributes
		DefaultMutableTreeNode eventAttributes = new DefaultMutableTreeNode(new TreeEntry(TREE_NODE_GLOBAL_EVENT_ATTRIBUTES));
		for (XAttribute attribute : log.getGlobalEventAttributes()) {
			eventAttributes.add(new DefaultMutableTreeNode(new TreeEntry(attribute.getKey())));
		}
		root.add(eventAttributes);

		// Build up list of Classifiers
		DefaultMutableTreeNode classifiers = new DefaultMutableTreeNode(new TreeEntry(TREE_NODE_CLASSIFIERS));
		for (XEventClassifier classifier : log.getClassifiers()) {
			DefaultMutableTreeNode nodeClassifier = new DefaultMutableTreeNode(new TreeEntry(classifier.name()));

			for (String att : classifier.getDefiningAttributeKeys()) {
				nodeClassifier.add(new DefaultMutableTreeNode(att));
			}

			classifiers.add(nodeClassifier);
		}
		root.add(classifiers);

		// Build up list of Attributes
		DefaultMutableTreeNode nodeAttributes = new DefaultMutableTreeNode(new TreeEntry(TREE_NODE_ATTRIBUTES));
		generateAttributesNode(log.getAttributes(), nodeAttributes);
		root.add(nodeAttributes);

		this.add(mdtreeScrollPane);

		// Adding the text area that shows the value for the selected metadata
		// node.

		// The text area that will show the value for the selected node.
		valueArea = new JTextArea();
		valueArea.setBackground(LogPreviewUI.COLOR_ENCLOSURE_BG);
		valueArea.setForeground(Color.WHITE);
		valueArea.setEditable(false);
		valueArea.setLineWrap(true);
		valueArea.setWrapStyleWord(true);

		// Create a scroll pane for the text area
		JScrollPane mdvalueScrollPane = new JScrollPane(valueArea);
		mdvalueScrollPane
				.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		mdvalueScrollPane.setOpaque(false);
		mdvalueScrollPane
				.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
		mdvalueScrollPane
				.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		JScrollBar vBar2 = mdvalueScrollPane.getVerticalScrollBar();
		vBar2.setUI(new SlickerScrollBarUI(vBar2,
				LogPreviewUI.COLOR_ENCLOSURE_BG, new Color(160, 160, 160),
				LogPreviewUI.COLOR_NON_FOCUS, 4, 12));
		JScrollBar hBar2 = mdvalueScrollPane.getHorizontalScrollBar();
		hBar2.setUI(new SlickerScrollBarUI(hBar2,
				LogPreviewUI.COLOR_ENCLOSURE_BG, new Color(160, 160, 160),
				LogPreviewUI.COLOR_NON_FOCUS, 4, 12));
		// Force a decent minimum width
		mdvalueScrollPane.setMinimumSize(new Dimension(230, 1000));
		mdvalueScrollPane.setPreferredSize(new Dimension(230, 1000));
		mdvalueScrollPane.setMaximumSize(new Dimension(230, 1000));

		this.add(mdvalueScrollPane);

		// Expand all top-level rows of the Metadata Tree.
		for (int i = 0; i < mdtree.getRowCount(); i++) {
			mdtree.expandRow(i);
		}
	}

	/**
	 * Update the text area to show the value for the current node.
	 * 
	 * @param value
	 *            The new value to be displayed in the text area.
	 * 
	 *            This may be extended to also show additional information.
	 */
	private void updateValueArea(String value) {
		valueArea.setText("Value:\n" + value);
	}
	
	private void updateValueArea(XAttribute attribute) {
		updateValueArea(attribute == null ? "" : attribute.toString());
	}

	private void updateValueArea() {
		valueArea
				.setText("There is no information available for the selected node.");
	}

	/**
	 * Recursively traverse all available attributes.
	 * 
	 * @param attributes
	 *            The attribute map from the log.
	 * @param nodeAttributes
	 *            The node under which all found attributes must be stored.
	 */
	private void generateAttributesNode(XAttributeMap attributes,
			DefaultMutableTreeNode nodeAttributes) {

		for (XAttribute attribute : attributes.values()) {
			DefaultMutableTreeNode nodeAttribute = new DefaultMutableTreeNode(new TreeEntry(attribute.getKey()));

			generateAttributesNode(attribute.getAttributes(), nodeAttribute);

			nodeAttributes.add(nodeAttribute);
		}
	}

	@Override
	public void valueChanged(TreeSelectionEvent e) {
		TreePath path = e.getPath();

		if (path == null) {
			return;
		}

		String rootComponent = (String) (((DefaultMutableTreeNode) path
				.getPathComponent(0)).getUserObject());
		if (TREE_ROOT_NODE.equals(rootComponent)) {

			// If the root node or one of the section nodes is selected, there
			// is no third component, so no information to show.
			if (path.getPathCount() <= 2) {
				updateValueArea();
				return;
			}

			// Handle requests from the section 'Attributes'. Show the
			// corresponding value in the 'value' text area (on the right).
			TreeEntry sectionComponent = (TreeEntry) (((DefaultMutableTreeNode) path
					.getPathComponent(1)).getUserObject());
			if (TREE_NODE_ATTRIBUTES.equals(sectionComponent.getKey())) {
				XAttribute attribute = getAttribute((Object[]) path.getPath(),
						2, log.getAttributes());

				updateValueArea(attribute);
			}
			else if(TREE_NODE_GLOBAL_EVENT_ATTRIBUTES.equals(sectionComponent.getKey())) {
				
				if(path.getPath().length > 3) {
					throw new IllegalArgumentException("Path is longer than expected.");
				}
				
				TreeEntry globalevent = (TreeEntry) ((DefaultMutableTreeNode)path.getLastPathComponent()).getUserObject();
				
				for(XAttribute attribute : log.getGlobalEventAttributes()) {
					if(globalevent.getKey().equals(attribute.getKey())) {
						updateValueArea(attribute);
						return;
					}
				}
				
				updateValueArea();
			}
			else if(TREE_NODE_GLOBAL_TRACE_ATTRIBUTES.equals(sectionComponent.getKey())) {

				if(path.getPath().length > 3) {
					throw new IllegalArgumentException("Path is longer than expected.");
				}
				
				TreeEntry globaltrace = (TreeEntry) ((DefaultMutableTreeNode)path.getLastPathComponent()).getUserObject();
				
				for(XAttribute attribute : log.getGlobalTraceAttributes()) {
					if(globaltrace.getKey().equals(attribute.getKey())) {
						updateValueArea(attribute);
						return;
					}
				}
				
				updateValueArea();
			}
			else if(TREE_NODE_EXTENSIONS.equals(sectionComponent.getKey())) {
				
				if(path.getPath().length > 3) {
					throw new IllegalArgumentException("Path is longer than expected.");
				}
				
				TreeEntry extensionnode = (TreeEntry) ((DefaultMutableTreeNode)path.getLastPathComponent()).getUserObject();
				
				for(XExtension extension : log.getExtensions()) {
					if(extensionnode.getKey().equals(extension.getName())) {
						updateValueArea(extension.getUri().toString());
						return;
					}
				}
				
				updateValueArea();
			}
			else {
				updateValueArea();
			}
		}
	}

	private XAttribute getAttribute(Object[] path, int position,
			XAttributeMap attributes) {
		TreeEntry pathComponent = (TreeEntry) ((DefaultMutableTreeNode) path[position])
				.getUserObject();
		XAttribute attribute = attributes.get(pathComponent.getKey());

		if (position == path.length - 1) {
			return attribute;
		}

		return getAttribute(path, position + 1, attribute.getAttributes());
	}
	
	private class TreeEntry {
		private String key;
		private String displaytext;
		
		public TreeEntry(String key) {
			this(key, key);
		}
		
		public TreeEntry(String key, String displaytext) {
			this.key = key;
			this.displaytext = displaytext;
		}
		
		public String getKey() {
			return this.key;
		}
		
		public String getDisplayText() {
			return this.displaytext;
		}
		
		public String toString() {
			return getDisplayText();
		}
	}
}
